import importlib

import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_almost_equal
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path

import act
from act.utils.data_utils import DatastreamParserARM as DatastreamParser
from act.utils.data_utils import convert_2d_to_1d

spec = importlib.util.find_spec('pyart')
if spec is not None:
    PYART_AVAILABLE = True
else:
    PYART_AVAILABLE = False


def test_add_in_nan():
    # Make a 1D array of 10 minute data
    time = np.arange('2019-01-01T01:00', '2019-01-01T01:10', dtype='datetime64[m]')
    time = time.astype('datetime64[ns]')
    time = np.delete(time, range(3, 8))
    data = np.linspace(0.0, 8.0, time.size)

    time_filled, data_filled = act.utils.add_in_nan(xr.DataArray(time), xr.DataArray(data))
    assert isinstance(time_filled, xr.core.dataarray.DataArray)
    assert isinstance(data_filled, xr.core.dataarray.DataArray)

    time_filled, data_filled = act.utils.add_in_nan(time, data)
    assert isinstance(time_filled, np.ndarray)
    assert isinstance(data_filled, np.ndarray)

    assert time_filled[3] == np.datetime64('2019-01-01T01:05:00')
    assert time_filled[4] == np.datetime64('2019-01-01T01:08:00')
    assert np.isnan(data_filled[3])
    assert data_filled[4] == 6.0

    time_filled, data_filled = act.utils.add_in_nan(time[0], data[0])
    assert time_filled == time[0]
    assert data_filled == data[0]

    # Check for multiple instances of missing data periods
    time = np.arange('2019-01-01T01:00', '2019-01-01T02:00', dtype='datetime64[m]')
    time = np.delete(time, range(3, 8))
    time = np.delete(time, range(33, 36))
    data = np.linspace(0.0, 10.0, time.size)

    time_filled, data_filled = act.utils.add_in_nan(time, data)
    assert time_filled.size == 54
    assert data_filled.size == 54
    index = np.where(time_filled == np.datetime64('2019-01-01T01:37'))[0]
    assert index[0] == 33
    assert np.isclose(data_filled[33], 6.27450)
    index = np.where(time_filled == np.datetime64('2019-01-01T01:38'))[0]
    assert index.size == 0
    index = np.where(time_filled == np.datetime64('2019-01-01T01:39'))[0]
    assert index[0] == 34
    assert np.isnan(data_filled[34])
    index = np.where(time_filled == np.datetime64('2019-01-01T01:40'))[0]
    assert index.size == 0

    # Test for 2D data
    time = np.arange('2019-01-01T01:00', '2019-01-01T02:00', dtype='datetime64[m]')
    data = np.random.random((len(time), 25))

    time = np.delete(time, range(3, 8))
    data = np.delete(data, range(3, 8), axis=0)
    time_filled, data_filled = act.utils.add_in_nan(time, data)

    assert np.count_nonzero(np.isnan(data_filled[3, :])) == 25
    assert len(time_filled) == len(time) + 2


def test_get_missing_value():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    missing = act.utils.data_utils.get_missing_value(
        ds, 'latent_heat_flux', use_FillValue=True, add_if_missing_in_ds=True
    )
    assert missing == -9999

    ds['latent_heat_flux'].attrs['missing_value'] = -9998
    missing = act.utils.data_utils.get_missing_value(ds, 'latent_heat_flux')
    assert missing == -9998


def test_convert_units():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    data = ds['soil_temp_1'].values
    in_units = ds['soil_temp_1'].attrs['units']
    r_data = act.utils.data_utils.convert_units(data, in_units, 'K')
    assert np.ceil(r_data[0]) == 285

    data = act.utils.data_utils.convert_units(r_data, 'K', 'C')
    assert np.ceil(data[0]) == 12

    with np.testing.assert_raises(ValueError):
        ds.utils.change_units()

    desired_unit = 'degF'
    skip_vars = [ii for ii in ds.data_vars if ii.startswith('qc_')]
    ds.utils.change_units(
        variables=None,
        desired_unit=desired_unit,
        skip_variables=skip_vars,
        skip_standard=True,
    )
    units = []
    for var_name in ds.data_vars:
        try:
            units.append(ds[var_name].attrs['units'])
        except KeyError:
            pass
    indices = [i for i, x in enumerate(units) if x == desired_unit]
    assert indices == [0, 2, 4, 6, 8, 32, 34, 36, 38, 40]

    var_name = 'home_signal_15'
    desired_unit = 'V'
    ds.utils.change_units(var_name, desired_unit, skip_variables='lat')
    assert ds[var_name].attrs['units'] == desired_unit

    var_names = ['home_signal_15', 'home_signal_30']
    ds.utils.change_units(var_names, desired_unit)
    for var_name in var_names:
        assert ds[var_name].attrs['units'] == desired_unit

    ds.close()
    del ds

    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_CEIL1)
    var_name = 'range'
    desired_unit = 'km'
    ds = ds.utils.change_units(var_name, desired_unit)
    assert ds[var_name].attrs['units'] == desired_unit
    assert np.isclose(np.sum(ds[var_name].values), 952.56, atol=0.01)

    ds.close()
    del ds

    # Test if exception or print statement is issued when an error occurs with units string
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    with np.testing.assert_raises(ValueError):
        ds.utils.change_units('home_signal_15', 'not_a_real_unit_string', raise_error=True)

    with np.testing.assert_raises(ValueError):
        ds.utils.change_units('not_a_real_variable_name', 'degC', raise_error=True)

    f = StringIO()
    var_name = 'home_signal_15'
    unit = 'not_a_real_unit_string'
    with redirect_stdout(f):
        ds.utils.change_units('home_signal_15', 'not_a_real_unit_string', verbose=True)
    s = f.getvalue()
    assert (
        s.strip()
        == f"Unable to convert '{var_name}' to units of '{unit}'. Skipping unit converstion for '{var_name}'."
    )
    ds.close()
    del ds


def test_ts_weighted_average():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET_WILDCARD)
    cf_ds = {
        'sgpmetE13.b1': {
            'variable': [
                'tbrg_precip_total',
                'org_precip_rate_mean',
                'pwd_precip_rate_mean_1min',
            ],
            'weight': [0.8, 0.15, 0.05],
            'ds': ds,
        }
    }
    data = act.utils.data_utils.ts_weighted_average(cf_ds)

    np.testing.assert_almost_equal(np.sum(data), 84.9, decimal=1)


def test_accum_precip():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET_WILDCARD)

    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total')
    dmax = round(np.nanmax(ds['tbrg_precip_total_accumulated']))
    assert np.isclose(dmax, 13.0, atol=0.01)

    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total', time_delta=60)
    dmax = round(np.nanmax(ds['tbrg_precip_total_accumulated']))
    assert np.isclose(dmax, 13.0, atol=0.01)

    ds['tbrg_precip_total'].attrs['units'] = 'mm/hr'
    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total')
    dmax = np.round(np.nanmax(ds['tbrg_precip_total_accumulated']), 2)
    assert np.isclose(dmax, 0.22, atol=0.01)


@pytest.mark.skipif(not PYART_AVAILABLE, reason='Py-ART is not installed.')
def test_create_pyart_obj():
    try:
        ds = act.io.mpl.read_sigma_mplv5(act.tests.EXAMPLE_SIGMA_MPLV5)
    except Exception:
        return

    radar = act.utils.create_pyart_obj(ds, range_var='range')
    variables = list(radar.fields)
    assert 'nrb_copol' in variables
    assert 'nrb_crosspol' in variables
    assert radar.sweep_start_ray_index['data'][-1] == 67
    assert radar.sweep_end_ray_index['data'][-1] == 101
    assert radar.fixed_angle['data'] == 2.0
    assert radar.scan_type == 'ppi'
    assert radar.sweep_mode['data'] == 'ppi'
    np.testing.assert_allclose(radar.sweep_number['data'][-3:], [1.0, 1.0, 1.0])
    np.testing.assert_allclose(radar.sweep_number['data'][0:3], [0.0, 0.0, 0.0])

    # coordinates
    np.testing.assert_allclose(radar.azimuth['data'][0:5], [-95.0, -92.5, -90.0, -87.5, -85.0])
    np.testing.assert_allclose(radar.elevation['data'][0:5], [2.0, 2.0, 2.0, 2.0, 2.0])
    np.testing.assert_allclose(
        radar.range['data'][0:5],
        [14.98962308, 44.96886923, 74.94811538, 104.92736153, 134.90660768],
    )
    gate_lat = radar.gate_latitude['data'][0, 0:5]
    gate_lon = radar.gate_longitude['data'][0, 0:5]
    gate_alt = radar.gate_altitude['data'][0, 0:5]
    np.testing.assert_allclose(
        gate_lat, [38.95293483, 38.95291135, 38.95288786, 38.95286437, 38.95284089]
    )
    np.testing.assert_allclose(
        gate_lon, [-76.8363515, -76.83669666, -76.83704182, -76.83738699, -76.83773215]
    )
    np.testing.assert_allclose(
        gate_alt, [62.84009906, 63.8864653, 64.93293721, 65.9795148, 67.02619806]
    )
    ds.close()
    del radar


def test_convert_to_potential_temp():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.convert_to_potential_temp(
        ds, temp_var_name, press_var_name=press_var_name
    )
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.001)
    temp = act.utils.data_utils.convert_to_potential_temp(
        temperature=ds[temp_var_name].values,
        pressure=ds[press_var_name].values,
        temp_var_units=ds[temp_var_name].attrs['units'],
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.0011)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=ds[temp_var_name].values,
            pressure=ds[press_var_name].values,
            temp_var_units=ds[temp_var_name].attrs['units'],
        )

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=ds[temp_var_name].values,
            pressure=ds[press_var_name].values,
            press_var_units=ds[press_var_name].attrs['units'],
        )


def test_height_adjusted_temperature():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_temperature(
        ds,
        temp_var_name,
        height_difference=100,
        height_units='m',
        press_var_name=press_var_name,
    )
    assert np.isclose(np.nansum(temp), -6834.291, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        ds, temp_var_name=temp_var_name, height_difference=-900, height_units='feet'
    )
    assert np.isclose(np.nansum(temp), -1904.7257, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        ds,
        temp_var_name,
        height_difference=-200,
        height_units='m',
        press_var_name=press_var_name,
        pressure=102.325,
        press_var_units='kPa',
    )
    assert np.isclose(np.nansum(temp), -2871.5435, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        height_difference=25.2,
        height_units='m',
        temperature=ds[temp_var_name].values,
        temp_var_units=ds[temp_var_name].attrs['units'],
        pressure=ds[press_var_name].values,
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), -5847.511, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_temperature(
            height_difference=25.2,
            height_units='m',
            temperature=ds[temp_var_name].values,
            temp_var_units=None,
            pressure=ds[press_var_name].values,
            press_var_units=ds[press_var_name].attrs['units'],
        )


def test_height_adjusted_pressure():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_pressure(
        ds=ds, press_var_name=press_var_name, height_difference=20, height_units='m'
    )
    assert np.isclose(np.nansum(temp), 142020.83, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_pressure(
        height_difference=-100,
        height_units='ft',
        pressure=ds[press_var_name].values,
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), 142877.69, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_pressure(
            height_difference=-100,
            height_units='ft',
            pressure=ds[press_var_name].values,
            press_var_units=None,
        )


def test_datastreamparser():
    test_values = [1234, 4321.0, True, ['sgpmetE13.b1'], ('sgpmetE13.b1',)]
    for test_value in test_values:
        pytest.raises(ValueError, DatastreamParser, test_value)

        fn_obj = DatastreamParser()
        pytest.raises(ValueError, fn_obj.set_datastream, test_values)

    fn_obj = DatastreamParser()
    assert fn_obj.site is None
    assert fn_obj.datastream_class is None
    assert fn_obj.facility is None
    assert fn_obj.level is None
    assert fn_obj.datastream is None
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None
    del fn_obj

    fn_obj = DatastreamParser('/data/sgp/sgpmetE13.b1/sgpmetE13.b1.20190501.024254.nc')
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility == 'E13'
    assert fn_obj.level == 'b1'
    assert fn_obj.datastream == 'sgpmetE13.b1'
    assert fn_obj.date == '20190501'
    assert fn_obj.time == '024254'
    assert fn_obj.ext == 'nc'

    fn_obj = DatastreamParser(Path('/data/sgp/sgpmetE13.b1/sgpmetE13.b1.20190501.024254.nc'))
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility == 'E13'
    assert fn_obj.level == 'b1'
    assert fn_obj.datastream == 'sgpmetE13.b1'
    assert fn_obj.date == '20190501'
    assert fn_obj.time == '024254'
    assert fn_obj.ext == 'nc'

    fn_obj.set_datastream('nsatwrC1.a0.19991230.233451.cdf')
    assert fn_obj.site == 'nsa'
    assert fn_obj.datastream_class == 'twr'
    assert fn_obj.facility == 'C1'
    assert fn_obj.level == 'a0'
    assert fn_obj.datastream == 'nsatwrC1.a0'
    assert fn_obj.date == '19991230'
    assert fn_obj.time == '233451'
    assert fn_obj.ext == 'cdf'

    fn_obj = DatastreamParser('nsaitscomplicatedX1.00.991230.2334.txt')
    assert fn_obj.site == 'nsa'
    assert fn_obj.datastream_class == 'itscomplicated'
    assert fn_obj.facility == 'X1'
    assert fn_obj.level == '00'
    assert fn_obj.datastream == 'nsaitscomplicatedX1.00'
    assert fn_obj.date == '991230'
    assert fn_obj.time == '2334'
    assert fn_obj.ext == 'txt'

    fn_obj = DatastreamParser('sgpmetE13.b1')
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility == 'E13'
    assert fn_obj.level == 'b1'
    assert fn_obj.datastream == 'sgpmetE13.b1'
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None

    fn_obj = DatastreamParser(Path('sgpmetE13.b1'))
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility == 'E13'
    assert fn_obj.level == 'b1'
    assert fn_obj.datastream == 'sgpmetE13.b1'
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None

    fn_obj = DatastreamParser('sgpmetE13')
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility == 'E13'
    assert fn_obj.level is None
    assert fn_obj.datastream is None
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None

    fn_obj = DatastreamParser('sgpmet')
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class == 'met'
    assert fn_obj.facility is None
    assert fn_obj.level is None
    assert fn_obj.datastream is None
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None

    fn_obj = DatastreamParser('sgp')
    assert fn_obj.site == 'sgp'
    assert fn_obj.datastream_class is None
    assert fn_obj.facility is None
    assert fn_obj.level is None
    assert fn_obj.datastream is None
    assert fn_obj.date is None
    assert fn_obj.time is None
    assert fn_obj.ext is None

    fn_obj = DatastreamParser(Path('zzzasoinfaoianasdfkansfaiZ99.s9.123456789.987654321.superlong'))
    assert fn_obj.site == 'zzz'
    assert fn_obj.datastream_class == 'asoinfaoianasdfkansfai'
    assert fn_obj.facility == 'Z99'
    assert fn_obj.level == 's9'
    assert fn_obj.datastream == 'zzzasoinfaoianasdfkansfaiZ99.s9'
    assert fn_obj.date == '123456789'
    assert fn_obj.time == '987654321'
    assert fn_obj.ext == 'superlong'

    values = [
        '',
        ' ',
        'sg',
        'SGP',
        'SGPMETE13.B1',
        Path('zzzasoinfaoianasdfkansfaiZ999.z1.123456789.987654321.superlong'),
        Path('/data/not/a/real/path/AsgpmetE13.b1.20190501.024254.nc'),
        '/data/not/a/real/path/AsgpmetE13.b1.20190501.024254.nc',
        'zzzasoinfaoianasdfkansfaiZ999.z1.123456789.987654321.superlong',
    ]
    for value in values:
        fn_obj = DatastreamParser(value)
        assert fn_obj.site is None
        assert fn_obj.datastream_class is None
        assert fn_obj.facility is None
        assert fn_obj.level is None
        assert fn_obj.datastream is None
        assert fn_obj.date is None
        assert fn_obj.time is None
        assert fn_obj.ext is None


def test_arm_site_location_search():
    # Test for many facilities
    test_dict_many = act.utils.arm_site_location_search(site_code='sgp', facility_code=None)
    assert len(test_dict_many) > 30
    assert list(test_dict_many)[0] == 'sgp A1'
    assert list(test_dict_many)[2] == 'sgp A3'

    assert_almost_equal(test_dict_many[list(test_dict_many)[0]]['latitude'], 37.843058)
    assert_almost_equal(test_dict_many[list(test_dict_many)[0]]['longitude'], -97.020569)
    assert_almost_equal(test_dict_many[list(test_dict_many)[2]]['latitude'], 37.626)
    assert_almost_equal(test_dict_many[list(test_dict_many)[2]]['longitude'], -96.882)

    # Test for one facility
    test_dict_one = act.utils.arm_site_location_search(site_code='sgp', facility_code='I5')
    assert len(test_dict_one) == 1
    assert list(test_dict_one)[0] == 'sgp I5'
    assert_almost_equal(test_dict_one[list(test_dict_one)[0]]['latitude'], 36.491178)
    assert_almost_equal(test_dict_one[list(test_dict_one)[0]]['longitude'], -97.593936)

    # Test for a facility with no latitude and longitude information
    test_dict_no_coord = act.utils.arm_site_location_search(site_code='sgp', facility_code='A6')
    assert list(test_dict_no_coord)[0] == 'sgp A6'
    assert test_dict_no_coord[list(test_dict_no_coord)[0]]['latitude'] is None
    assert test_dict_no_coord[list(test_dict_no_coord)[0]]['longitude'] is None

    # Test for another site
    test_dict_nsa = act.utils.arm_site_location_search(site_code='nsa', facility_code=None)
    assert len(test_dict_nsa) > 5
    assert list(test_dict_nsa)[0] == 'nsa C1'
    assert test_dict_nsa[list(test_dict_nsa)[0]]['latitude'] == 71.323
    assert test_dict_nsa[list(test_dict_nsa)[0]]['longitude'] == -156.615


def test_calculate_percentages():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_AOSACSM)
    fields = ['sulfate', 'ammonium', 'nitrate', 'chloride']
    time = '2023-04-20T03:49:45.000000000'
    time_slice = ('2023-04-20T17:38:20.000000000', '2023-04-20T20:29:47.000000000')
    threshold = 0.0

    # Without threshold, chloride has invalid negative values so
    # percentages will be incorrect. Check if warning is created
    with pytest.warns(UserWarning) as record:
        act.utils.calculate_percentages(
            ds, fields, time='2023-04-20T03:49:45.000000000', threshold=None
        )
        if not record:
            pytest.fail("Expected a warning for invalid data.")

    # Test with threshold and singular time
    percentages = act.utils.calculate_percentages(ds, fields, time=time, threshold=threshold)
    assert 'sulfate' in percentages.keys()
    assert 'chloride' in percentages.keys()
    assert np.round(percentages["sulfate"], 3) == 66.125
    assert np.round(percentages["chloride"], 3) == 0.539

    # Test with sliced time
    percentages = act.utils.calculate_percentages(ds, fields, time_slice=time_slice, threshold=0.0)
    assert np.round(percentages["sulfate"], 3) == 68.342
    assert np.round(percentages["chloride"], 3) == 1.042

    # Run on all times and check if warning exists.
    with pytest.warns(UserWarning) as record:
        percentages = act.utils.calculate_percentages(ds, fields, threshold=0.0)
        assert np.round(percentages["sulfate"], 3) == 66.373
        assert np.round(percentages["chloride"], 3) == 0.915
        if not record:
            pytest.fail("Expected a warning for using all times.")


def test_convert_2d_to_1d():
    # Create a  sample dataset
    data = np.array([[1, 2], [3, 4], [5, 6]])
    ds = xr.Dataset(
        {'var': (('time', 'level'), data)}, coords={'time': [0, 1, 2], 'level': [10, 20]}
    )
    ds['level'].attrs['units'] = 'm'

    # Run the function
    result = convert_2d_to_1d(ds, parse='level')

    # Check the results
    assert 'var_level_0' in result
    assert 'var_level_1' in result
    np.testing.assert_array_equal(result['var_level_0'].values, [1, 3, 5])
    np.testing.assert_array_equal(result['var_level_1'].values, [2, 4, 6])

    # Run the function with use_dim_value_in_name=True
    result = convert_2d_to_1d(ds, parse='level', use_dim_value_in_name=True)

    # Check the results
    assert 'var_level_10m' in result
    assert 'var_level_20m' in result
    np.testing.assert_array_equal(result['var_level_10m'].values, [1, 3, 5])
    np.testing.assert_array_equal(result['var_level_20m'].values, [2, 4, 6])

    # Run the function with custom labels
    result = convert_2d_to_1d(ds, parse='level', dim_labels=['low', 'high'])

    # Check the results
    assert 'var_low' in result
    assert 'var_high' in result
    np.testing.assert_array_equal(result['var_low'].values, [1, 3, 5])
    np.testing.assert_array_equal(result['var_high'].values, [2, 4, 6])

    # Create a sample dataset
    data = np.array([[1], [3], [5]])
    ds = xr.Dataset({'var': (('time', 'level'), data)}, coords={'time': [0, 1, 2], 'level': [10]})

    # Run the function with keep_name_if_one=True
    result = convert_2d_to_1d(ds, parse='level', keep_name_if_one=True)

    # Check the results
    assert 'var' in result
    np.testing.assert_array_equal(result['var'].values, [1, 3, 5])



"""
Skew-T plot of a sounding
-------------------------

This example shows how to make a Skew-T plot from a sounding
and calculate stability indicies.

Author: Maxwell Grover

"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt
import numpy as np
import xarray as xr

import act

# Make sure attributes are retained
xr.set_options(keep_attrs=True)

# Read data
filename_sonde = DATASETS.fetch('twpsondewnpnC3.b1.20060121.231600.custom.cdf')
sonde_ds = act.io.arm.read_arm_netcdf(filename_sonde)


# Calculate stability indicies
sonde_ds = act.retrievals.calculate_stability_indicies(
    sonde_ds, temp_name='tdry', td_name='dp', p_name='pres'
)

# Plot the stability index values on the plot
variables = [
    'lifted_index',
    'surface_based_cape',
    'surface_based_cin',
    'most_unstable_cape',
    'most_unstable_cin',
    'lifted_condensation_level_temperature',
    'lifted_condensation_level_pressure',
]


# Add a helper function which will format the text
def format_variable(variable, rounding_digits=2):
    """Format a sounding variable to displayed on a single line"""
    return f'{variable}: {np.round(sonde_ds[variable], rounding_digits).values} {sonde_ds[variable].units}'


# Setup the plot
skewt = act.plotting.SkewTDisplay(sonde_ds, figsize=(12, 8))

# Add the stability indices
ax = skewt.axes[0]
props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
for i in range(len(variables)):
    ax.text(
        0.05,
        (0.98 - (0.05 * i)),
        format_variable(variables[i]),
        transform=ax.transAxes,
        fontsize=10,
        verticalalignment='top',
        bbox=props,
    )

# Add data
skewt.plot_from_u_and_v('u_wind', 'v_wind', 'pres', 'tdry', 'dp', shade_cin=False)

sonde_ds.close()
plt.show()



""" Module for Distribution Plotting. """

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import pandas as pd

from ..utils import datetime_utils as dt_utils, calculate_percentages
from .plot import Display


class DistributionDisplay(Display):
    """
    This class is used to make  distribution related plots. It is inherited from Display
    and therefore contains all of Display's attributes and methods.

    Examples
    --------
    To create a DistributionDisplay with 3 rows, simply do:

    .. code-block:: python

        ds = act.io.read_arm_netcdf(the_file)
        disp = act.plotting.DistsributionDisplay(ds, subplot_shape=(3,), figsize=(15, 5))

    The DistributionDisplay constructor takes in the same keyword arguments as
    plt.subplots. For more information on the plt.subplots keyword arguments,
    see the `matplotlib documentation
    <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.subplots.html>`_.
    If no subplot_shape is provided, then no figure or axis will be created
    until add_subplots or plots is called.

    """

    def __init__(self, ds, subplot_shape=(1,), ds_name=None, **kwargs):
        super().__init__(ds, subplot_shape, ds_name, **kwargs)

    def set_xrng(self, xrng, subplot_index=(0,)):
        """
        Sets the x range of the plot.

        Parameters
        ----------
        xrng : 2 number array
            The x limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError('set_xrng requires the plot to be displayed.')

        if not hasattr(self, 'xrng') and len(self.axes.shape) == 2:
            self.xrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2), dtype='datetime64[D]')
        elif not hasattr(self, 'xrng') and len(self.axes.shape) == 1:
            self.xrng = np.zeros((self.axes.shape[0], 2), dtype='datetime64[D]')

        self.axes[subplot_index].set_xlim(xrng)
        self.xrng[subplot_index, :] = np.array(xrng)

    def set_yrng(self, yrng, subplot_index=(0,)):
        """
        Sets the y range of the plot.

        Parameters
        ----------
        yrng : 2 number array
            The y limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError('set_yrng requires the plot to be displayed.')

        if not hasattr(self, 'yrng') and len(self.axes.shape) == 2:
            self.yrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2))
        elif not hasattr(self, 'yrng') and len(self.axes.shape) == 1:
            self.yrng = np.zeros((self.axes.shape[0], 2))

        if yrng[0] == yrng[1]:
            yrng[1] = yrng[1] + 1

        self.axes[subplot_index].set_ylim(yrng)
        self.yrng[subplot_index, :] = yrng

    def _get_data(self, dsname, fields):
        if isinstance(fields, str):
            fields = [fields]
        return self._ds[dsname][fields].dropna('time')

    def plot_stacked_bar(
        self,
        field,
        dsname=None,
        bins=10,
        sortby_field=None,
        sortby_bins=None,
        subplot_index=(0,),
        set_title=None,
        density=False,
        hist_kwargs=dict(),
        **kwargs,
    ):
        """
        This procedure will plot a stacked bar graph of a histogram.

        Parameters
        ----------
        field : str
            The name of the field to take the histogram of.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        bins : array-like or int
            The histogram bin boundaries to use. If not specified, numpy's
            default 10 is used.
        sortby_field : str or None
            Set this option to a field name in order to sort the histograms
            by a given field parameter. For example, one can sort histograms of CO2
            concentration by temperature.
        sortby_bins : array-like or None
            The bins to sort the histograms by.
        subplot_index : tuple
            The subplot index to place the plot in
        set_title : str
            The title of the plot.
        density : bool
            Set to True to plot a p.d.f. instead of a frequency histogram.
        hist_kwargs : dict
            Additional keyword arguments to pass to numpy histogram.

        Other keyword arguments will be passed into :func:`matplotlib.pyplot.bar`.

        Returns
        -------
        return_dict : dict
            A dictionary containing the plot axis handle, bin boundaries, and
            generated histogram.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                + 'or more datasets in the TimeSeriesDisplay '
                + 'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        if sortby_field is not None:
            ds = self._get_data(dsname, [field, sortby_field])
            xdata, ydata = ds[field], ds[sortby_field]
        else:
            xdata = self._get_data(dsname, field)[field]

        if 'units' in xdata.attrs:
            xtitle = ''.join(['(', xdata.attrs['units'], ')'])
        else:
            xtitle = field

        if sortby_bins is None and sortby_field is not None:
            # We will defaut the y direction to have the same # of bins as x
            if isinstance(bins, int):
                n_bins = bins
            else:
                n_bins = len(bins)
            sortby_bins = np.linspace(ydata.values.min(), ydata.values.max(), n_bins)

        # Get the current plotting axis
        if self.fig is None:
            self.fig = plt.figure()
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        if sortby_field is not None:
            if 'units' in ydata.attrs:
                ytitle = ''.join(['(', ydata.attrs['units'], ')'])
            else:
                ytitle = field
            my_hist, x_bins, y_bins = np.histogram2d(
                xdata.values.flatten(),
                ydata.values.flatten(),
                density=density,
                bins=[bins, sortby_bins],
                **hist_kwargs,
            )
            x_inds = (x_bins[:-1] + x_bins[1:]) / 2.0
            self.axes[subplot_index].bar(
                x_inds,
                my_hist[:, 0].flatten(),
                label=(str(y_bins[0]) + ' to ' + str(y_bins[1])),
                **kwargs,
            )
            for i in range(1, len(y_bins) - 1):
                self.axes[subplot_index].bar(
                    x_inds,
                    my_hist[:, i].flatten(),
                    bottom=my_hist[:, i - 1],
                    label=(str(y_bins[i]) + ' to ' + str(y_bins[i + 1])),
                    **kwargs,
                )
            self.axes[subplot_index].legend()
        else:
            my_hist, bins = np.histogram(
                xdata.values.flatten(), bins=bins, density=density, **hist_kwargs
            )
            x_inds = (bins[:-1] + bins[1:]) / 2.0
            self.axes[subplot_index].bar(x_inds, my_hist)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    field,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        self.axes[subplot_index].set_title(set_title)
        self.axes[subplot_index].set_ylabel('count')
        self.axes[subplot_index].set_xlabel(xtitle)

        return_dict = {}
        return_dict['plot_handle'] = self.axes[subplot_index]
        if 'x_bins' in locals():
            return_dict['x_bins'] = x_bins
            return_dict['y_bins'] = y_bins
        else:
            return_dict['bins'] = bins
        return_dict['histogram'] = my_hist

        return return_dict

    def plot_size_distribution(
        self, field, bins, time=None, dsname=None, subplot_index=(0,), set_title=None, **kwargs
    ):
        """
        This procedure plots a stairstep plot of a size distribution. This is
        useful for plotting size distributions and waveforms.

        Parameters
        ----------
        field : str
            The name of the field to plot the spectrum from.
        bins : str or array-like
            The name of the field that stores the bins for the spectra.
        time : none or datetime
            If None, spectra to plot will be automatically determined.
            Otherwise, specify this field for the time period to plot.
        dsname : str
            The name of the Dataset to plot. Set to None to have
            ACT automatically determine this.
        subplot_index : tuple
            The subplot index to place the plot in.
        set_title : str or None
            Use this to set the title.

        Additional keyword arguments will be passed into :func:`matplotlib.pyplot.step`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle referring to the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                + 'or more datasets in the TimeSeriesDisplay '
                + 'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        xdata = self._get_data(dsname, field)[field]

        if isinstance(bins, str):
            bins = self._ds[dsname][bins]
        else:
            bins = xr.DataArray(bins)

        if 'units' in bins.attrs:
            xtitle = ''.join(['(', bins.attrs['units'], ')'])
        else:
            xtitle = 'Bin #'

        if 'units' in xdata.attrs:
            ytitle = ''.join(['(', xdata.attrs['units'], ')'])
        else:
            ytitle = field

        if len(xdata.dims) > 1 and time is None:
            raise ValueError(
                'Input data has more than one dimension, ' + 'you must specify a time to plot!'
            )
        elif len(xdata.dims) > 1:
            xdata = xdata.sel(time=time, method='nearest')

        if len(bins.dims) > 1 or len(bins.values) != len(xdata.values):
            raise ValueError(
                'Bins must be a one dimensional field whose '
                + 'length is equal to the field length!'
            )

        # Get the current plotting axis
        if self.fig is None:
            self.fig = plt.figure()
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    field,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
            if time is not None:
                t = pd.Timestamp(time)
                set_title += ''.join(
                    [' at ', ':'.join([str(t.hour), str(t.minute), str(t.second)])]
                )
        self.axes[subplot_index].set_title(set_title)
        self.axes[subplot_index].step(bins.values, xdata.values, **kwargs)
        self.axes[subplot_index].set_xlabel(xtitle)
        self.axes[subplot_index].set_ylabel(ytitle)

        return self.axes[subplot_index]

    def plot_stairstep(
        self,
        field,
        dsname=None,
        bins=10,
        sortby_field=None,
        sortby_bins=None,
        subplot_index=(0,),
        set_title=None,
        density=False,
        hist_kwargs=dict(),
        **kwargs,
    ):
        """
        This procedure will plot a stairstep plot of a histogram.

        Parameters
        ----------
        field : str
            The name of the field to take the histogram of.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        bins : array-like or int
            The histogram bin boundaries to use. If not specified, numpy's
            default 10 is used.
        sortby_field : str or None
            Set this option to a field name in order to sort the histograms
            by a given field parameter. For example, one can sort histograms of CO2
            concentration by temperature.
        sortby_bins : array-like or None
            The bins to sort the histograms by.
        subplot_index : tuple
            The subplot index to place the plot in.
        set_title : str
            The title of the plot.
        density : bool
            Set to True to plot a p.d.f. instead of a frequency histogram.
        hist_kwargs : dict
            Additional keyword arguments to pass to numpy histogram.

        Other keyword arguments will be passed into :func:`matplotlib.pyplot.step`.

        Returns
        -------
        return_dict : dict
             A dictionary containing the plot axis handle, bin boundaries, and
             generated histogram.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                + 'or more datasets in the TimeSeriesDisplay '
                + 'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        xdata = self._get_data(dsname, field)[field]

        if 'units' in xdata.attrs:
            xtitle = ''.join(['(', xdata.attrs['units'], ')'])
        else:
            xtitle = field

        if sortby_field is not None:
            ydata = self._ds[dsname][sortby_field]

        if sortby_bins is None and sortby_field is not None:
            if isinstance(bins, int):
                n_bins = bins
            else:
                n_bins = len(bins)
            # We will defaut the y direction to have the same # of bins as x
            sortby_bins = np.linspace(ydata.values.min(), ydata.values.max(), n_bins)

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        if sortby_field is not None:
            if 'units' in ydata.attrs:
                ytitle = ''.join(['(', ydata.attrs['units'], ')'])
            else:
                ytitle = field
            my_hist, x_bins, y_bins = np.histogram2d(
                xdata.values.flatten(),
                ydata.values.flatten(),
                density=density,
                bins=[bins, sortby_bins],
                **hist_kwargs,
            )
            x_inds = (x_bins[:-1] + x_bins[1:]) / 2.0
            self.axes[subplot_index].step(
                x_inds,
                my_hist[:, 0].flatten(),
                label=(str(y_bins[0]) + ' to ' + str(y_bins[1])),
                **kwargs,
            )
            for i in range(1, len(y_bins) - 1):
                self.axes[subplot_index].step(
                    x_inds,
                    my_hist[:, i].flatten(),
                    label=(str(y_bins[i]) + ' to ' + str(y_bins[i + 1])),
                    **kwargs,
                )
            self.axes[subplot_index].legend()
        else:
            my_hist, bins = np.histogram(
                xdata.values.flatten(), bins=bins, density=density, **hist_kwargs
            )

            x_inds = (bins[:-1] + bins[1:]) / 2.0
            self.axes[subplot_index].step(x_inds, my_hist, **kwargs)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    field,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        self.axes[subplot_index].set_title(set_title)
        self.axes[subplot_index].set_ylabel('count')
        self.axes[subplot_index].set_xlabel(xtitle)

        return_dict = {}
        return_dict['plot_handle'] = self.axes[subplot_index]
        if 'x_bins' in locals():
            return_dict['x_bins'] = x_bins
            return_dict['y_bins'] = y_bins
        else:
            return_dict['bins'] = bins
        return_dict['histogram'] = my_hist

        return return_dict

    def plot_heatmap(
        self,
        x_field,
        y_field,
        dsname=None,
        x_bins=None,
        y_bins=None,
        subplot_index=(0,),
        set_title=None,
        density=False,
        set_shading='auto',
        hist_kwargs=dict(),
        threshold=None,
        **kwargs,
    ):
        """
        This procedure will plot a heatmap of a histogram from 2 variables.

        Parameters
        ----------
        x_field : str
            The name of the field to take the histogram of on the X axis.
        y_field : str
            The name of the field to take the histogram of on the Y axis.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        x_bins : array-like, int, or None
            The histogram bin boundaries to use for the variable on the X axis.
            Set to None to use numpy's default boundaries.
            If an int, will indicate the number of bins to use
        y_bins : array-like, int, or None
            The histogram bin boundaries to use for the variable on the Y axis.
            Set to None to use numpy's default boundaries.
            If an int, will indicate the number of bins to use
        subplot_index : tuple
            The subplot index to place the plot in
        set_title : str
            The title of the plot.
        density : bool
            Set to True to plot a p.d.f. instead of a frequency histogram.
        set_shading : string
            Option to to set the matplotlib.pcolormesh shading parameter.
            Default to 'auto'
        threshold : float
            Value on which to threshold the histogram results for plotting.
            Setting to 0 will ensure that all 0 values are removed from the plot
            making it easier to distringuish between 0 and low values
        hist_kwargs : Additional keyword arguments to pass to numpy histogram.

        Other keyword arguments will be passed into :func:`matplotlib.pyplot.pcolormesh`.

        Returns
        -------
        return_dict : dict
            A dictionary containing the plot axis handle, bin boundaries, and
            generated histogram.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        ds = self._get_data(dsname, [x_field, y_field])
        xdata, ydata = ds[x_field], ds[y_field]

        if 'units' in xdata.attrs:
            xtitle = ''.join(['(', xdata.attrs['units'], ')'])
        else:
            xtitle = x_field

        if x_bins is not None and isinstance(x_bins, int):
            x_bins = np.linspace(xdata.values.min(), xdata.values.max(), x_bins)

        if y_bins is not None and isinstance(x_bins, int):
            y_bins = np.linspace(ydata.values.min(), ydata.values.max(), y_bins)

        if x_bins is not None and y_bins is None:
            # We will defaut the y direction to have the same # of bins as x
            y_bins = np.linspace(ydata.values.min(), ydata.values.max(), len(x_bins))

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        if 'units' in ydata.attrs:
            ytitle = ''.join(['(', ydata.attrs['units'], ')'])
        else:
            ytitle = y_field

        if x_bins is None:
            my_hist, x_bins, y_bins = np.histogram2d(
                xdata.values.flatten(), ydata.values.flatten(), density=density, **hist_kwargs
            )
        else:
            my_hist, x_bins, y_bins = np.histogram2d(
                xdata.values.flatten(),
                ydata.values.flatten(),
                density=density,
                bins=[x_bins, y_bins],
                **hist_kwargs,
            )
        # Adding in the ability to threshold the heatmaps
        if threshold is not None:
            my_hist[my_hist <= threshold] = np.nan

        x_inds = (x_bins[:-1] + x_bins[1:]) / 2.0
        y_inds = (y_bins[:-1] + y_bins[1:]) / 2.0
        xi, yi = np.meshgrid(x_inds, y_inds, indexing='ij')
        mesh = self.axes[subplot_index].pcolormesh(xi, yi, my_hist, shading=set_shading, **kwargs)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        self.axes[subplot_index].set_title(set_title)
        self.axes[subplot_index].set_ylabel(ytitle)
        self.axes[subplot_index].set_xlabel(xtitle)
        self.add_colorbar(mesh, title='count', subplot_index=subplot_index)

        return_dict = {}
        return_dict['plot_handle'] = self.axes[subplot_index]
        return_dict['x_bins'] = x_bins
        return_dict['y_bins'] = y_bins
        return_dict['histogram'] = my_hist

        return return_dict

    def set_ratio_line(self, subplot_index=(0,)):
        """
        Sets the 1:1 ratio line.

        Parameters
        ----------
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError('set_ratio_line requires the plot to be displayed.')
        # Define the xticks of the figure
        xlims = self.axes[subplot_index].get_xticks()
        ratio = np.linspace(xlims, xlims[-1])
        self.axes[subplot_index].plot(ratio, ratio, 'k--')

    def plot_scatter(
        self,
        x_field,
        y_field,
        m_field=None,
        dsname=None,
        cbar_label=None,
        set_title=None,
        subplot_index=(0,),
        **kwargs,
    ):
        """
        This procedure will produce a scatter plot from 2 variables.

        Parameters
        ----------
        x_field : str
            The name of the field to display on the X axis.
        y_field : str
            The name of the field to display on the Y axis.
        m_field : str
            The name of the field to display on the markers.
        cbar_label : str
            The desired name to plot for the colorbar
        set_title : str
            The desired title for the plot.
            Default title is created from the datastream.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        subplot_index : tuple
            The subplot index to place the plot in

        Other keyword arguments will be passed into :func:`matplotlib.pyplot.scatter`.

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        if m_field is None:
            mdata = None
            ds = self._get_data(dsname, [x_field, y_field])
            xdata, ydata = ds[x_field], ds[y_field]
        else:
            ds = self._get_data(dsname, [x_field, y_field, m_field])
            xdata, ydata, mdata = ds[x_field], ds[y_field], ds[m_field]

        # Define the x-axis label. If units are avaiable, plot.
        if 'units' in xdata.attrs:
            xtitle = x_field + ''.join([' (', xdata.attrs['units'], ')'])
        else:
            xtitle = x_field

        # Define the y-axis label. If units are available, plot
        if 'units' in ydata.attrs:
            ytitle = y_field + ''.join([' (', ydata.attrs['units'], ')'])
        else:
            ytitle = y_field

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        # Define the axes for the figure
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Display the scatter plot, pass keyword args for unspecified attributes
        scc = self.axes[subplot_index].scatter(xdata, ydata, c=mdata, **kwargs)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )

        # Check to see if a colorbar label was set
        if mdata is not None:
            if cbar_label is None:
                # Define the y-axis label. If units are available, plot
                if 'units' in mdata.attrs:
                    ztitle = m_field + ''.join([' (', mdata.attrs['units'], ')'])
                else:
                    ztitle = m_field
            else:
                ztitle = cbar_label
            # Plot the colorbar
            cbar = plt.colorbar(scc)
            cbar.ax.set_ylabel(ztitle)

        # Define the axe title, x-axis label, y-axis label
        self.axes[subplot_index].set_title(set_title)
        self.axes[subplot_index].set_ylabel(ytitle)
        self.axes[subplot_index].set_xlabel(xtitle)

        return self.axes[subplot_index]

    def plot_violin(
        self,
        field,
        positions=None,
        dsname=None,
        orientation='vertical',
        showmeans=True,
        showmedians=True,
        showextrema=True,
        subplot_index=(0,),
        set_title=None,
        **kwargs,
    ):
        """
        This procedure will produce a violin plot for the selected
        field (or fields).

        Parameters
        ----------
        field : str or list
            The name of the field (or fields) to display on the X axis.
        positions : array-like, Default: None
            The positions of the ticks along dependent axis.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        orientation : str
            Orientation of the violin plot. Options are 'vertical' and 'horizontal'.
            Default is vertical.
        showmeans : Boolean; Default: False
            If True, will display the mean of the datastream.
        showmedians : Boolean; Default: False
            If True, will display the medium of the datastream.
        showextrema: Boolean; Default: False
            If True, will display the extremes of the datastream.
        subplot_index : tuple
            The subplot index to place the plot in
        set_title : str
            The title of the plot.

        Other keyword arguments will be passed into :func:`matplotlib.pyplot.violinplot`.

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        if dsname is None:
            dsname = list(self._ds.keys())[0]

        ds = self._get_data(dsname, field)
        ndata = ds[field]

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        # Define the axes for the figure
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Define the axe label. If units are avaiable, plot.
        if 'units' in ndata.attrs:
            axtitle = field + ''.join([' (', ndata.attrs['units'], ')'])
        else:
            axtitle = field

        # Display the scatter plot, pass keyword args for unspecified attributes
        scc = self.axes[subplot_index].violinplot(
            ndata,
            positions=positions,
            orientation=orientation,
            showmeans=showmeans,
            showmedians=showmedians,
            showextrema=showextrema,
            **kwargs,
        )
        if showmeans is True:
            scc['cmeans'].set_edgecolor('red')
            scc['cmeans'].set_label('mean')
        if showmedians is True:
            scc['cmedians'].set_edgecolor('black')
            scc['cmedians'].set_label('median')
        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )

        # Define the axe title, x-axis label, y-axis label
        self.axes[subplot_index].set_title(set_title)
        if orientation == 'vertical':
            self.axes[subplot_index].set_ylabel(axtitle)
            if positions is None:
                self.axes[subplot_index].set_xticks([])
        else:
            self.axes[subplot_index].set_xlabel(axtitle)
            if positions is None:
                self.axes[subplot_index].set_yticks([])

        return self.axes[subplot_index]

    def plot_pie_chart(
        self,
        fields,
        time=None,
        time_slice=None,
        threshold=None,
        fill_value=0.0,
        dsname=None,
        subplot_index=(0,),
        set_title=None,
        autopct='%1.1f%%',
        **kwargs,
    ):
        """
        This procedure will produce a pie chart for the selected fields.

        Parameters
        ----------
        fields : list
            The list of fields to calculate percentages on for the pie chart.
        time : datetime
            A single datetime to be passed into the act.utils.calculate percentages function
            if desired. Default is None and all data will be included.
        time_slice : tuple
            A tuple of two datetimes to grab all data between those two datetimes for
            act.utils.calculate_percentages. Default is None and all data will be included.
        threshold : float
            Threshold in which anything below will be considered invalid.
            Default is None.
        fill_value : float
            Fill value for invalid data. Only used if a threshold is provided.
        dsname : str or None
            The name of the datastream the field is contained in. Set
            to None to let ACT automatically determine this.
        subplot_index : tuple
            The subplot index to place the plot in
        set_title : str
            The title of the plot.
        autopct : str
            Format string for the percentages. Default is float with one
            decimal place. If this parameter is set to None, no percentage
            string values are displayed.
        **kwargs : keywords
            Keywords to pass through to :func:`matplotlib.pyplot.pie`.

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                + 'or more datasets in the DistributionDisplay '
                + 'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get the current plotting axis
        if self.fig is None:
            self.fig = plt.figure()
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        self.axes[subplot_index].set_title(set_title)

        percentages = calculate_percentages(
            self._ds[dsname],
            fields,
            time=time,
            time_slice=time_slice,
            threshold=threshold,
            fill_value=fill_value,
        )

        self.axes[subplot_index].pie(
            [percentages[field] for field in percentages.keys()],
            labels=percentages.keys(),
            autopct=autopct,
            **kwargs,
        )
        return self.axes[subplot_index]



import matplotlib
import pytest

import act
from act.plotting import XSectionDisplay
from act.tests import sample_files

try:
    import cartopy  # noqa

    CARTOPY_AVAILABLE = True
except ImportError:
    CARTOPY_AVAILABLE = False

matplotlib.use('Agg')


def test_xsection_errors():
    ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_CEIL1)

    display = XSectionDisplay(ds, figsize=(10, 8), subplot_shape=(2,))
    display.axes = None
    pytest.raises(RuntimeError, display.set_yrng, [0, 10])
    pytest.raises(RuntimeError, display.set_xrng, [-40, 40])

    display = XSectionDisplay(ds, figsize=(10, 8), subplot_shape=(1,))
    pytest.raises(
        RuntimeError, display.plot_xsection, 'backscatter', x='time', cmap='HomeyerRainbow'
    )

    ds.close()
    matplotlib.pyplot.close(fig=display.fig)


@pytest.mark.mpl_image_compare(tolerance=20)
def test_xsection_plot():
    visst_ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_CEIL1)

    xsection = XSectionDisplay(visst_ds, figsize=(10, 8))
    xsection.plot_xsection('backscatter', x='time', y='range', cmap='coolwarm', vmin=0, vmax=320)
    visst_ds.close()

    try:
        return xsection.fig
    finally:
        matplotlib.pyplot.close(xsection.fig)


@pytest.mark.skipif(not CARTOPY_AVAILABLE, reason='Cartopy is not installed.')
@pytest.mark.mpl_image_compare(tolerance=10)
def test_xsection_plot_map():
    radar_ds = act.io.arm.read_arm_netcdf(
        sample_files.EXAMPLE_VISST, combine='nested', concat_dim='time'
    )
    try:
        with pytest.warns(
            UserWarning,
            match="Could not discern datastreamname and dict or tuple were not provided. Using defaultname of act_datastream!",
        ):
            xsection = XSectionDisplay(radar_ds, figsize=(15, 8))
        xsection.plot_xsection_map(
            'ir_temperature',
            vmin=220,
            vmax=300,
            cmap='Greys',
            x='longitude',
            y='latitude',
            isel_kwargs={'time': 0},
        )
        radar_ds.close()
        try:
            return xsection.fig
        finally:
            matplotlib.pyplot.close(xsection.fig)
    except Exception:
        pass


@pytest.mark.skipif(not CARTOPY_AVAILABLE, reason='Cartopy is not installed.')
@pytest.mark.mpl_image_compare(tolerance=10)
def test_xsection_plot_map_background():
    radar_ds = act.io.arm.read_arm_netcdf(
        sample_files.EXAMPLE_VISST, combine='nested', concat_dim='time'
    )
    try:
        with pytest.warns(
            UserWarning,
            match="Could not discern datastreamname and dict or tuple were not provided. Using defaultname of act_datastream!",
        ):
            xsection = XSectionDisplay(radar_ds, figsize=(15, 8))
        xsection.plot_xsection_map(
            'ir_temperature',
            vmin=220,
            vmax=300,
            cmap='Greys',
            x='longitude',
            y='latitude',
            background=True,
            isel_kwargs={'time': 0},
        )
        radar_ds.close()
        try:
            return xsection.fig
        finally:
            matplotlib.pyplot.close(xsection.fig)
    except Exception:
        pass



"""
This module contains procedures for working with QC information
and for applying tests to data.

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        'add_supplemental_qc',
        'arm',
        'bsrn_tests',
        'clean',
        'comparison_tests',
        'qcfilter',
        'qc_summary',
        'qctests',
        'radiometer_tests',
        'sp2',
    ],
    submod_attrs={
        'arm': ['add_dqr_to_qc', 'print_dqr'],
        'qcfilter': [
            'QCFilter',
            'parse_bit',
            'set_bit',
            'unset_bit',
        ],
        'qc_summary': ['QCSummary'],
        'qctests': [
            'QCTests',
        ],
        'clean': ['CleanDataset'],
        'radiometer_tests': [
            'fft_shading_test',
            'fft_shading_test_process',
        ],
        'bsrn_tests': ['QCTests'],
        'comparison_tests': ['QCTests'],
        'add_supplemental_qc': ['read_yaml_supplemental_qc'],
        'sp2': ['SP2ParticleCriteria', 'get_waveform_statistics'],
    },
)



"""
Simple plot of 2D data
----------------------

This is an example of how to download and
plot ceiliometer data from the SGP site
over Oklahoma.

"""

import os

from arm_test_data import DATASETS
import matplotlib.pyplot as plt

import act

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# If the username and token are not set, use the existing sample file
if username is None or token is None or len(username) == 0 or len(token) == 0:
    filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
    ceil_ds = act.io.arm.read_arm_netcdf(filename_ceil, engine='netcdf4')
else:
    # Example to show how easy it is to download ARM data if a username/token are set
    results = act.discovery.download_arm_data(
        username, token, 'sgpceilC1.b1', '2022-01-14', '2022-01-19'
    )
    ceil_ds = act.io.arm.read_arm_netcdf(results)

# Adjust ceilometer data for plotting
ceil_ds = act.corrections.ceil.correct_ceil(ceil_ds, -9999.0)

# Plot up ceilometer backscatter using HomeyerRainbow CVD friendly colormap
# The same could be done with keyword 'cmap='HomeyerRainbow'
display = act.plotting.TimeSeriesDisplay(ceil_ds, subplot_shape=(1,), figsize=(15, 5))
display.plot('backscatter', subplot_index=(0,), cvd_friendly=True)
plt.show()



"""
Create a dataset to mimic ARM file formats
------------------------------------------
Example shows how to create a dataset from an ARM DOD.
This will enable users to create files that mimic ARM
files, making for easier use across the community.

Author: Adam Theisen

"""

import act

# Create an empty dataset using an ARM DOD
ds = act.io.arm.create_ds_from_arm_dod('vdis.b1', {'time': 1440}, scalar_fill_dim='time')

# Print out the xarray dataset to see that it's empty
print(ds)

# The user could populate this through a number of ways
# and that's best left up to the user on how to do it.
# If one has an existing dataset, a mapping of variable
# names is sometimes the easiest way

# Let's look at some variable attributes
# These can be updated and it would be up to the
# user to ensure these tests are being applied
# and are appropriately set in the cooresponding QC variable
print(ds['num_drops'].attrs)

# Next, let's print out the global attribuets
print(ds.attrs)

# Add additional attributes or append to existing
# if they are needed using a dictionary
atts = {
    'command_line': 'python  plot_create_arm_ds.py',
    'process_version': '1.2.3',
    'history': 'Processed with Jupyter Workbench',
    'random': '1234253sdgfadf',
}
for a in atts:
    if a in ds.attrs:
        ds.attrs[a] += atts[a]
    else:
        ds.attrs[a] = atts[a]
    # Print out the attribute
    print(a, ds.attrs[a])

# Write data out to netcdf
ds.to_netcdf('./sgpvdisX1.b1.20230101.000000.nc')

# If one wants to clean up the dataset to better match CF standards
# the following can be done as well
ds.write.write_netcdf(cf_compliant=True, path='./sgpvdisX1.b1.20230101.000000.cf')



"""
Plotting Baseline Surface Radiation Network (BSRN) QC Flags
-----------------------------------------------------------

Simple example for applying BSRN QC and
plotting the data and the corresponding QC flags
using colorblind friendly colors.
https://bsrn.awi.de/data/quality-checks/

Author: Ken Kehoe

"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in data and convert from ARM QC standard to CF QC standard
filename_brs = DATASETS.fetch('sgpbrsC1.b1.20190705.000000.cdf')
ds = act.io.arm.read_arm_netcdf(filename_brs, cleanup_qc=True)

# Creat Plot Display and plot data including embedded QC from data file
variable = 'down_short_hemisp'
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot radiation data in top plot
display.plot(variable, subplot_index=(0,), day_night_background=True, cvd_friendly=True)

# Plot ancillary QC data in bottom plot
display.qc_flag_block_plot(variable, subplot_index=(1,), cvd_friendly=True)
plt.show()

# Add initial BSRN QC tests to ancillary QC varialbles. Use defualts for
# test set to Physicall Possible and use_dask.
ds.qcfilter.bsrn_limits_test(
    gbl_SW_dn_name='down_short_hemisp',
    glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
    direct_normal_SW_dn_name='short_direct_normal',
    glb_SW_up_name='up_short_hemisp',
    glb_LW_dn_name='down_long_hemisp_shaded',
    glb_LW_up_name='up_long_hemisp',
)

# Add initial BSRN QC tests to ancillary QC varialbles. Use defualts for
# test set to Extremely Rare" and to use Dask processing.
ds.qcfilter.bsrn_limits_test(
    test='Extremely Rare',
    gbl_SW_dn_name='down_short_hemisp',
    glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
    direct_normal_SW_dn_name='short_direct_normal',
    glb_SW_up_name='up_short_hemisp',
    glb_LW_dn_name='down_long_hemisp_shaded',
    glb_LW_up_name='up_long_hemisp',
    use_dask=True,
)

# Add comparison BSRN QC tests to ancillary QC varialbles. Request two of the possible
# comparison tests.
ds.qcfilter.bsrn_comparison_tests(
    ['Global over Sum SW Ratio', 'Diffuse Ratio'],
    gbl_SW_dn_name='down_short_hemisp',
    glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
    direct_normal_SW_dn_name='short_direct_normal',
)

# Add K-tests QC to ancillary QC variables.
ds.qcfilter.normalized_rradiance_test(
    [
        'Clearness index',
        'Upper total transmittance',
        'Upper direct transmittance',
        'Upper diffuse transmittance',
    ],
    dni='short_direct_normal',
    dhi='down_short_hemisp',
    ghi='down_short_diffuse_hemisp',
)

# Creat Plot Display and plot data including embedded QC from data file
variable = 'down_short_hemisp'
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot radiation data in top plot. Add QC information to top plot.
display.plot(
    variable,
    subplot_index=(0,),
    day_night_background=True,
    assessment_overplot=True,
    cvd_friendly=True,
)

# Plot ancillary QC data in bottom plot
display.qc_flag_block_plot(variable, subplot_index=(1,), cvd_friendly=True)
plt.show()



"""
Script for downloading data from ARM's Live Data Webservice

"""

import json
import os
from datetime import timedelta
import requests
import textwrap
import warnings

try:
    from urllib.request import Request, urlopen
except ImportError:
    from urllib import Request, urlopen

from act.utils import date_parser


def download_arm_data(username, token, datastream, startdate, enddate, time=None, output=None):
    """
    This tool will help users utilize the ARM Live Data Webservice to download
    ARM data.

    Parameters
    ----------
    username : str
        The username to use for logging into the ADC archive.
    token : str
        The access token for accessing the ADC archive.
    datastream : str
        The name of the datastream to acquire.
    startdate : str
        The start date of the data to acquire. Formats accepted are
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY, YYYYMMDD, YYYY/MM/DD or
        any of the previous formats with THH:MM:SS added onto the end
        (ex. 2020-09-15T12:00:00).
    enddate : str
        The end date of the data to acquire. Formats accepted are
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY, YYYYMMDD or YYYY/MM/DD, or
        any of the previous formats with THH:MM:SS added onto the end
        (ex. 2020-09-15T13:00:00).
    time: str or None
        The specific time. Format is HHMMSS. Set to None to download all files
        in the given date interval.
    output : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as *datastream* to place
        the files in.

    Returns
    -------
    files : list
        Returns list of files retrieved

    Notes
    -----
    This programmatic interface allows users to query and automate
    machine-to-machine downloads of ARM data. This tool uses a REST URL and
    specific parameters (saveData, query), user ID and access token, a
    datastream name, a start date, and an end date, and data files matching
    the criteria will be returned to the user and downloaded.

    By using this web service, users can setup cron jobs and automatically
    download data from /data/archive into their workspace. This will also
    eliminate the manual step of following a link in an email to download data.
    All other data files, which are not on the spinning
    disk (on HPSS), will have to go through the regular ordering process.
    More information about this REST API and tools can be found on `ARM Live
    <https://adc.arm.gov/armlive/#scripts>`_.

    To login/register for an access token click `here
    <https://adc.arm.gov/armlive/>`_.

    Author: Michael Giansiracusa

    Examples
    --------
    This code will download the netCDF files from the sgpmetE13.b1 datastream
    and place them in a directory named sgpmetE13.b1. The data from 14 Jan to
    20 Jan 2017 will be downloaded. Replace *userName* and *XXXXXXXXXXXXXXXX*
    with your username and token for ARM Data Discovery. See the Notes for
    information on how to obtain a username and token.

    .. code-block:: python

        act.discovery.download_data(
            "userName", "XXXXXXXXXXXXXXXX", "sgpmetE13.b1", "2017-01-14", "2017-01-20"
        )

    """
    # default start and end are empty
    start, end = '', ''
    # start and end strings for query_url are constructed
    # if the arguments were provided
    if startdate:
        start_datetime = date_parser(startdate, return_datetime=True)
        start = start_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        start = f'&start={start}'
    if enddate:
        end_datetime = date_parser(enddate, return_datetime=True)
        # If the start and end date are the same, and a day to the end date
        if start_datetime == end_datetime:
            end_datetime += timedelta(hours=23, minutes=59, seconds=59)
        end = end_datetime.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        end = f'&end={end}'
    # build the url to query the web service using the arguments provided
    query_url = (
        'https://adc.arm.gov/armlive/data/query?' + 'user={0}&ds={1}{2}{3}&wt=json'
    ).format(':'.join([username, token]), datastream, start, end)

    headers = {
        'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36',
    }

    req = Request(query_url, None, headers)
    # get url response, read the body of the message,
    # and decode from bytes type to utf-8 string
    response_body = urlopen(req).read().decode('utf-8')
    # if the response is an html doc, then there was an error with the user
    if response_body[1:14] == '!DOCTYPE html':
        raise ConnectionRefusedError('Error with user. Check username or token.')

    # parse into json object
    response_body_json = json.loads(response_body)

    # construct output directory
    if output:
        # output files to directory specified
        output_dir = os.path.join(output)
    else:
        # if no folder given, add datastream folder
        # to current working dir to prevent file mix-up
        output_dir = os.path.join(os.getcwd(), datastream)

    # not testing, response is successful and files were returned
    if response_body_json is None:
        print('ARM Data Live Webservice does not appear to be functioning')
        return []

    num_files = len(response_body_json['files'])
    file_names = []
    if response_body_json['status'] == 'success' and num_files > 0:
        for fname in response_body_json['files']:
            if time is not None:
                if time not in fname:
                    continue
            # construct link to web service saveData function
            save_data_url = (
                'https://adc.arm.gov/armlive/data/' + 'saveData?user={0}&file={1}'
            ).format(':'.join([username, token]), fname)
            req_save = Request(save_data_url, None, headers)
            output_file = os.path.join(output_dir, fname)
            # make directory if it doesn't exist
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
            # create file and write bytes to file
            with open(output_file, 'wb') as open_bytes_file:
                data = urlopen(req_save).read()
                if 'This data file is not available' in str(data):
                    print(fname + ' is not available for download')
                    continue
                else:
                    print(f'[DOWNLOADING] {fname}')
                    open_bytes_file.write(data)
            file_names.append(output_file)
        # Get ARM DOI and print it out
        doi = get_arm_doi(
            datastream, start_datetime.strftime('%Y-%m-%d'), end_datetime.strftime('%Y-%m-%d')
        )
        print('\nIf you use these data to prepare a publication, please cite:\n')
        print(textwrap.fill(doi, width=80))
        print('')
    else:
        print(
            'No files returned or url status error.\n'
            'Check datastream name, start, and end date. \n'
        )
        if datastream.split('.')[-1] == 'a0':
            print(
                '---------------------ATTENTION--------------------\n',
                'ARM does not generally make a0-level data discoverable\n',
                'Please reach out through the ARM Help link to request access',
            )

    return file_names


def download_arm_data_mod(username, token, file_list, variables=None, filetype='cdf', output=None):
    """
    This function allows for downloading ARM data from ARM's data discovery mod API:
    https://adc.arm.gov/armlive/

    This function will then provide a concatenated file of the file list in csv or
    cdf file format with all variables or user provided variables.

    Parameters
    ----------
    username : str
        ARM username.
    token : str
        ARM user token.
    file_list : list
        A list of files by datetime of the same datastream to concatenate.
    variables : list
        A list of desired variables that exist in the datastream.
        Default is None and will provide all variables.
    file_type : str
        File type to download file as. Valid options are 'csv' and 'cdf'.
        Default is 'cdf'
    output : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as filename to place
        the files in.

    Returns
    -------
    filename : str
        Concatenated filename
    """
    if variables is not None:
        variables_str = ",".join(variables)
        url = f'https://adc.arm.gov/armlive/mod?user={username}:{token}&variables={variables_str}&wt={filetype}'
    else:
        url = f'https://adc.arm.gov/armlive/mod?user={username}:{token}&wt={filetype}'
    headers = {'Content-Type': 'application/json'}
    response = requests.get(url, headers=headers, data=json.dumps(file_list))

    if response.status_code == 200:
        # Save the response content to a file
        filename = response.headers.get('Content-Disposition').split('filename=')[1]
        filename = filename.replace('"', '')

        if output is not None:
            # output files to directory specified
            output_dir = os.path.join(output)
        else:
            # if no folder given, set current directory as folder
            output_dir = os.getcwd()

        # make directory if it doesn't exist and its a user provided folder.
        if not os.path.isdir(output_dir) and output is not None:
            os.makedirs(output_dir)

        output_file = os.path.join(output_dir, filename)

        with open(output_file, 'wb') as file:
            file.write(response.content)

        print(f"File saved as {filename}")
        return filename
    else:
        warnings.warn(
            'No files returned or url status error.\n'
            'Check if username, token, file list are valid. \n',
            UserWarning,
        )
        return


def get_arm_doi(datastream, startdate, enddate):
    """
    This function will return a citation with DOI, if available, for specified
    datastream and date range

    Parameters
    ----------
    datastream : str
        The name of the datastream to get a DOI for.  This must be ARM standard names
    startdate : str
        Start date for the citation in the format YY-MM-DD
    enddate : str
        End date for the citation in the format YY-MM-DD

    Returns
    -------
    doi : str
        Returns the citation as a string

    """
    headers = {
        'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36',
    }
    # Get the DOI information
    doi_url = (
        'https://adc.arm.gov/citationservice/citation/datastream?id='
        + datastream
        + '&citationType=apa'
    )
    doi_url += '&startDate=' + startdate
    doi_url += '&endDate=' + enddate
    try:
        doi = requests.get(url=doi_url, headers=headers)
    except ValueError as err:
        return "Webservice potentially down or arguments are not valid: " + err

    if len(doi.text) > 0:
        doi = doi.json()['citation']
    else:
        doi = 'Please check your arguments. No DOI Found'

    return doi



"""
Plot Present Weather Code
--------------------------

Plot the Present Weather Code on Precipitation Accumulation

Author: Joe O'Brien

"""

from arm_test_data import DATASETS
from matplotlib.dates import DateFormatter
from matplotlib.dates import num2date
import matplotlib.pyplot as plt

import act

# Read the MET data into an xarray dataset
filename_met = DATASETS.fetch('gucmetM1.b1.20230301.000000.cdf')
ds = act.io.read_arm_netcdf(filename_met)

# Decode the Present Weather Codes
# Pass it to the function to decode it along with the variable name
ds = act.utils.inst_utils.decode_present_weather(ds, variable='pwd_pw_code_inst')

# Calculate Precipitation Accumulation
pre_accum = act.utils.accumulate_precip(
    ds.where(ds.qc_tbrg_precip_total == 0), "tbrg_precip_total"
).tbrg_precip_total_accumulated.compute()

# Add the Precipitation Accum to the MET DataSet
ds['tbrg_accum'] = pre_accum

# Create a matplotlib figure
fig, ax = plt.subplots(1, 1, figsize=(10, 10))
# Adjust subplot width
fig.subplots_adjust(hspace=0.09)

# Create ACT display
display = act.plotting.TimeSeriesDisplay(ds)

# Define the Date/Time Format
date_form = DateFormatter("%H%M UTC")

# Assign the ACT display object to the matplotlib figure subplot
display.assign_to_figure_axis(fig, ax)
# Datastream Names are needed for plotting!
display.plot('tbrg_accum', label='TBRG Accumualated Precip')

# Add a day/night background
display.day_night_background()

# Update axe information and formatting!
ax.set_ylabel('Precipitation Accumulation [mm]')
# Add a title
ax.set_title('MET Tipping Bucket Rain Gauge - Crested Butte, CO')
# Define the x-axis format
ax.xaxis.set_major_formatter(date_form)
# Define the x-axis label
ax.set_xlabel('Time [UTC]')
# Gridlines are helpful
ax.grid(True)

# Grab the X-ticks (and convert to datetime objects) to plot location of PWD codes
xticks = display.axes[0].get_xticks()
ndates = [num2date(x) for x in xticks]

# Grab the PWD codes associated with those ticks
ncode = [
    ds['pwd_pw_code_inst_decoded'].sel(time=x.replace(tzinfo=None), method='nearest').data.tolist()
    for x in ndates
]
pwd_code = ['\n'.join(x.split(' ')) if len(x) > 20 else x for x in ncode]

# Display these select PWD codes as vertical texts along the x-axis
# Define the minimum y-axis tick mark for plotting
ymin = display.axes[0].get_yticks()[0]

# Plot the PWD code
for i, key in enumerate(xticks):
    ax.text(key, ymin, pwd_code[i], rotation=90, va='center')

plt.subplots_adjust(bottom=0.20)

plt.show()



import numpy as np
import xarray as xr

import act


def test_correct_ceil():
    # Make a fake ARM dataset to test with, just an array with 1e-7 for half
    # of it
    fake_data = 10 * np.ones((300, 20))
    fake_data[:, 10:] = -1
    arm_ds = {}
    arm_ds['backscatter'] = xr.DataArray(fake_data)
    arm_ds = act.corrections.ceil.correct_ceil(arm_ds)
    assert np.all(arm_ds['backscatter'].data[:, 10:] == -7)
    assert np.all(arm_ds['backscatter'].data[:, 1:10] == 1)

    arm_ds['backscatter'].attrs['units'] = 'dummy'
    arm_ds = act.corrections.ceil.correct_ceil(arm_ds)
    assert arm_ds['backscatter'].units == 'log(dummy)'



"""
This module contains I/O operations for loading csv files.

"""

import pathlib

import pandas as pd

from act.io.arm import check_arm_standards


def read_csv(
    filename, sep=',', engine='python', column_names=None, skipfooter=0, ignore_index=True, **kwargs
):
    """
    Returns an `xarray.Dataset` with stored data and metadata from user-defined
    query of CSV files.

    Parameters
    ----------
    filenames : str or list
        Name of file(s) to read.
    sep : str
        The separator between columns in the csv file.
    column_names : list or None
        The list of column names in the csv file.
    verbose : bool
        If true, will print if a file is not found.
    ignore_index : bool
         Keyword for pandas concat function.  If True, do not use the index
         values along the concatenation axis. The resulting axis will be labeled
         0, …, n - 1. This is useful if you are concatenating datasets where the
         concatenation axis does not have meaningful indexing information. Note
         the index values on the other axes are still respected in the join.

    Additional keyword arguments will be passed into pandas.read_csv.

    Returns
    -------
    ds : xarray.Dataset
        ACT Xarray dataset. Will be None if the file is not found.

    Examples
    --------
    This example will load the example sounding data used for unit testing:

    .. code-block:: python

        import act

        ds = act.io.csv.read(act.tests.sample_files.EXAMPLE_CSV_WILDCARD)

    """

    # Convert to string if filename is a pathlib or not a list
    if isinstance(filename, (pathlib.PurePath, str)):
        filename = [str(filename)]

    if isinstance(filename, list) and isinstance(filename[0], pathlib.PurePath):
        filename = [str(ii) for ii in filename]

    # Read data using pandas read_csv one file at a time and append to
    # list. Then concatinate the list into one pandas dataframe.
    li = []
    for fl in filename:
        df = pd.read_csv(
            fl, sep=sep, names=column_names, skipfooter=skipfooter, engine=engine, **kwargs
        )
        li.append(df)

    if len(li) == 1:
        df = li[0]
    else:
        df = pd.concat(li, axis=0, ignore_index=ignore_index)

    # Set Coordinates if there's a variable date_time
    if 'date_time' in df:
        df.date_time = df.date_time.astype('datetime64[ns]')
        df.time = df.date_time
        df = df.set_index('time')

    # Convert to xarray DataSet
    ds = df.to_xarray()

    # Set additional variables
    # Since we cannot assume a standard naming convention setting
    # file_date and file_time to the first time in the file
    x_coord = ds.coords.to_index().values[0]
    if isinstance(x_coord, str):
        x_coord_dt = pd.to_datetime(x_coord)
        ds.attrs['_file_dates'] = x_coord_dt.strftime('%Y%m%d')
        ds.attrs['_file_times'] = x_coord_dt.strftime('%H%M%S')

    # Check for standard ARM datastream name, if none, assume the file is ARM
    # standard format.
    is_arm_file_flag = check_arm_standards(ds)
    if is_arm_file_flag == 0:
        ds.attrs['_datastream'] = '.'.join(filename[0].split('/')[-1].split('.')[0:2])

    # Add additional attributes, site, standards flag, etc...
    ds.attrs['_site'] = str(ds.attrs['_datastream'])[0:3]
    ds.attrs['_arm_standards_flag'] = is_arm_file_flag

    return ds



"""
Working with and expanding embedded quality control variables
-------------------------------------------------------------

This is an example of how to use existing or create new quality
control variables and extend the quality control flagging. The
anicllary quality control variable can be expanded by integrating
external Data Quality Reports, adding additional generic ACT tests,
instrument specific ACT tests, or reading a configuration file of
known failures to clean up the data variable.

"""

import os

from arm_test_data import DATASETS
import matplotlib.pyplot as plt

import act

# Place your username and token here for use with ARM Live service
# https://adc.arm.gov/armlive/
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# We can use the ACT module for downloading data from the ARM web service
if username is None or token is None or len(username) == 0 or len(token) == 0:
    results = DATASETS.fetch('sgpmfrsr7nchE11.b1.20210329.070000.nc')
else:
    results = act.discovery.download_arm_data(
        username, token, 'sgpmfrsr7nchE11.b1', '2021-03-29', '2021-03-29'
    )
print(results)

# Let's plot up some data to see what we're working with. For this example, we'll use
# diffuse_hemisp_narrowband_filter4. The data files uses an ancillary quality control variable
# named with the same value prepened with 'qc_'. To read less data into memory we can tell
# the reader to only read in the variables we plan to use with the keep_variables keyword.
variable = 'diffuse_hemisp_narrowband_filter4'
qc_variable = 'qc_' + variable

# Next up is to read the file into an xarray object using the ACT reader. We then can print out a
# listing of everything in the object. Also, call the cleanup method on the object by setting
# the cleanup_qc keyword. This will convert the quality control variable from the ARM stanard
# to Climate and Forecast standard used internally for all the quality control calls.
keep_vars = [variable, qc_variable, 'lat', 'lon']
ds = act.io.arm.read_arm_netcdf(results, keep_variables=keep_vars, cleanup_qc=True)
print(ds)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()

# Now lets remove some of these outliers by modifying the data in the Dataset based on
# the test results stored in the ancillary quality control variable.
# By default the ancillary quality control variable is removed after appying the test
# results, but we are going to use the del_qc_var to keep in Dataset so it
# can be used with additional tests later.
ds.qcfilter.datafilter(variable, rm_tests=[2, 3], del_qc_var=False)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()

# Since the embedded QC is not removing all the outliers, let's check to see if there are any
# Data Quality Reports (DQR) using ARMs DQR Webservice. The great thing is, that ACT has codes
# for working with this webservice.
# In this example, we can see that there's a DQR for a shadowband misalignment and we can
# find out more information by looking at the actual DQR:
# https://adc.arm.gov/ArchiveServices/DQRService?dqrid=D210405.5

# Query the ARM DQR Webservice and update the ancillary quality control variable to
# contain a new test using information from the DQR.
ds = act.qc.arm.add_dqr_to_qc(ds, variable=variable)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()

# ACT has a number of additional QC tests that could be applied to the data. For this next
# example, let's apply a new maximum test. We are also
# going to filter the data based on this new test and plot up the results.

# Add a new maximum tests
ds.qcfilter.add_greater_test(variable, 0.4, test_meaning='New maximum tests limit')

# Filter that test out
ds.qcfilter.datafilter(variable, rm_tests=5, del_qc_var=False)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()

# ACT has a growing library of instrument specific tests such as the fast-fourier transform
# test to detect shading which was adapted from Alexandrov et al 2007. The adaption is that
# it is applied in a moving window style approach.

# Apply test
ds = act.qc.fft_shading_test(ds, variable=variable)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()

# The orginal embedded quality control variable plus additional tests we applied
# to the data did a good job removing incorrect data, but we can manually clean up the
# data a little more. By inspecting the data we can extract the time ranges where
# data is incorrect and create a new test using those time ranges. Instead of hardcoding
# those into the program we can write them to a YAML file for use in other programs or
# to give to other users.
# There is a file in the same directory called sgpmfrsr7nchE11.b1.yaml with times of
# incorrect or suspect values that can be read and applied to the Dataset.
from act.qc.add_supplemental_qc import apply_supplemental_qc  # noqa

apply_supplemental_qc(ds, 'sgpmfrsr7nchE11.b1.yaml')

# We can apply or reapply the data filter on the variable in the Dataset to change
# the data values failing tests to NaN by passing a list of test numbers we want
# to use. In this case we are not going to apply the DQR test (number 4) so we leave
# that number out of the list.
ds.qcfilter.datafilter(variable, rm_tests=[2, 3, 5, 6, 7, 8], del_qc_var=False)

# Create a plotting display object with 2 plots
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot up the diffuse variable in the first plot
display.plot(variable, subplot_index=(0,), day_night_background=True)

# Plot up the QC variable in the second plot
display.qc_flag_block_plot(variable, subplot_index=(1,))

plt.show()



import numpy as np

from act.io.arm import read_arm_netcdf
from act.qc.radiometer_tests import fft_shading_test
from act.tests import EXAMPLE_MFRSR


def test_fft_shading_test():
    ds = read_arm_netcdf(EXAMPLE_MFRSR)
    ds.clean.cleanup()
    ds = fft_shading_test(ds)
    qc_data = ds['qc_diffuse_hemisp_narrowband_filter4']
    assert np.nansum(qc_data.values) == 7164



import pytest


def pytest_addoption(parser):
    parser.addoption("--runbig", action="store_true", default=False, help="Run big tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "big: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runbig"):
        # --runbig given in cli: do not skip big tests
        return
    skip_big = pytest.mark.skip(reason="need --runbig option to run")
    for item in items:
        if "big" in item.keywords:
            item.add_marker(skip_big)



"""
Function for downloading data from NOAA PSL Profiler Network

"""
from datetime import datetime
import pandas as pd
import os

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


def download_noaa_psl_data(
    site=None, instrument=None, startdate=None, enddate=None, hour=None, output=None
):
    """
    Function to download data from the NOAA PSL Profiler Network Data Library
    https://psl.noaa.gov/data/obs/datadisplay/

    Parameters
    ----------
    site : str
        3 letter NOAA site identifier. Required variable
    instrument : str
        Name of the dataset to download.  Options currently include (name prior to -). Required variable
            'Parsivel' - Parsivel disdrometer data
            'Pressure', 'Datalogger', 'Net Radiation', 'Temp/RH', 'Solar Radiation' - Surface meteorology/radiation data
            'Tipping Bucket', 'TBRG', 'Wind Speed', 'Wind Direction' - Surface meteorology/radiation data
            'Wind Speed and Direction' - Surface meteorology/radiation data
            'GpsTrimble' - GPS Trimble water vapor data
            'Radar S-band Moment' - 3 GHz Precipitation Profiler moment data
            'Radar S-band Bright Band' - 3 GHz Precipitation Profiler bright band data
            '449RWP Bright Band' - 449 MHz Wind Profiler bright band data
            '449RWP Wind' - 449 MHz Wind Profiler wind data
            '449RWP Sub-Hour Wind' - 449 MHz Wind Profiler sub-hourly wind data
            '449RWP Sub-Hour Temp' - 449 MHz Wind Profiler sub-hourly temperature data
            '915RWP Wind' - 915 MHz Wind Profiler wind data
            '915RWP Temp' - 915 MHz Wind Profiler temperature data
            '915RWP Sub-Hour Wind' - 915 MHz Wind Profiler sub-hourly wind data
            '915WP Sub-Hour Temp' - 915 MHz Wind Profiler sub-hourly temperature data
            'Radar FMCW Moment' - FMCW Radar moments data
            'Radar FMCW Bright Band' - FMCW Radar bright band data
    startdate : str
        The start date of the data to acquire. Format is YYYYMMDD. Required variable
    enddate : str
        The end date of the data to acquire. Format is YYYYMMDD
    hour : str
        Two digit hour of file to dowload if wanting a specific time
    output : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as *datastream* to place
        the files in.

    Returns
    -------
    files : list
        Returns list of files retrieved

    """

    if (site is None) or (instrument is None) or (startdate is None):
        raise ValueError('site, instrument, and startdate need to be set')

    datastream = site + '_' + instrument.replace(' ', '_')
    # Convert dates to day of year (doy) for NOAA folder structure
    s_doy = datetime.strptime(startdate, '%Y%m%d').timetuple().tm_yday
    year = datetime.strptime(startdate, '%Y%m%d').year
    if enddate is None:
        enddate = startdate
    e_doy = datetime.strptime(enddate, '%Y%m%d').timetuple().tm_yday

    # Set base URL
    url = 'https://downloads.psl.noaa.gov/psd2/data/realtime/'

    # Set list of strings that all point to the surface meteorology dataset
    met_ds = [
        'Pressure',
        'Datalogger',
        'Net Radiation',
        'Temp/RH',
        'Solar Radiation',
        'Tipping Bucket',
        'TBRG',
        'Wind Speed',
        'Wind Direction',
        'Wind Speed and Direction',
    ]

    # Add to the url depending on which instrument is requested
    if 'Parsivel' in instrument:
        url += 'DisdrometerParsivel/Stats/'
    elif any([d in instrument for d in met_ds]):
        url += 'CsiDatalogger/SurfaceMet/'
    elif 'GpsTrimble' in instrument:
        url += 'GpsTrimble/WaterVapor/'
    elif 'Radar S-band Moment' in instrument:
        url += 'Radar3000/PopMoments/'
    elif 'Radar S-band Bright Band' in instrument:
        url += 'Radar3000/BrightBand/'
    elif '449RWP Bright Band' in instrument:
        url += 'Radar449/BrightBand/'
    elif '449RWP Wind' in instrument:
        url += 'Radar449/WwWind/'
    elif '449RWP Sub-Hour Wind' in instrument:
        url += 'Radar449/WwWindSubHourly/'
    elif '449RWP Sub-Hour Temp' in instrument:
        url += 'Radar449/WwTempSubHourly/'
    elif '915RWP Wind' in instrument:
        url += 'Radar915/WwWind/'
    elif '915RWP Temp' in instrument:
        url += 'Radar915/WwTemp/'
    elif '915RWP Sub-Hour Wind' in instrument:
        url += 'Radar915/WwWindSubHourly/'
    elif '915RWP Sub-Hour Temp' in instrument:
        url += 'Radar915/WwTempSubHourly/'
    elif 'Radar FMCW Moment' in instrument:
        url += 'RadarFMCW/PopMoments/'
    elif 'Radar FMCW Bright Band' in instrument:
        url += 'RadarFMCW/BrightBand/'
    else:
        raise ValueError('Instrument not supported')

    # Construct output directory
    if output:
        # Output files to directory specified
        output_dir = os.path.join(output)
    else:
        # If no folder given, add datastream folder
        # to current working dir to prevent file mix-up
        output_dir = os.path.join(os.getcwd(), datastream)

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # Set up doy ranges, taking into account changes for a new year
    prev_doy = 0
    if e_doy < s_doy:
        r = list(range(s_doy, 366)) + list(range(1, e_doy + 1))
    else:
        r = list(range(s_doy, e_doy + 1))

    # Set filename variable to return
    filenames = []

    # Loop through each doy in range
    for doy in r:
        # if the previous day is greater than current, assume a new year
        # i.e. 365 -> 001
        if prev_doy > doy:
            year += 1
        # Add site, year, and 3-digit day to url
        new_url = url + site + '/' + str(year) + '/' + str(doy).zfill(3) + '/'

        # User pandas to get a list of filenames to download
        # Exclude the first and last records which are "parent directory" and "nan"
        files = pd.read_html(new_url, skiprows=[1])[0]['Name']
        files = list(files[1:-1])

        # Write each file out to a file with same name as online
        for f in files:
            if hour is not None:
                if (str(doy).zfill(3) + str(hour)) not in f and (
                    str(doy).zfill(3) + '.' + str(hour)
                ) not in f:
                    continue
            output_file = os.path.join(output_dir, f)
            try:
                print('Downloading ' + f)
                with open(output_file, 'wb') as open_bytes_file:
                    open_bytes_file.write(urlopen(new_url + f).read())
                filenames.append(output_file)
            except Exception:
                pass
        prev_doy = doy

    return filenames



import matplotlib
import pytest

import act
from act.plotting import ContourDisplay
from act.tests import sample_files

matplotlib.use('Agg')


@pytest.mark.mpl_image_compare(tolerance=10)
def test_contour():
    files = sample_files.EXAMPLE_MET_CONTOUR
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        ds = act.io.arm.read_arm_netcdf(f)
        data.update({f: ds})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50, contour='contour', cmap='viridis')
    display.plot_vectors_from_spd_dir(
        fields=wind_fields, time=time, mesh=True, grid_delta=(0.1, 0.1)
    )
    display.plot_station(fields=station_fields, time=time, markersize=7, color='red')

    try:
        return display.fig
    finally:
        matplotlib.pyplot.close(display.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_contour_stamp():
    files = sample_files.EXAMPLE_STAMP_WILDCARD
    test = {}
    stamp_fields = {}
    time = '2020-01-01T00:00:00.000000000'
    for f in files:
        ds = f.split('/')[-1]
        nc_ds = act.io.arm.read_arm_netcdf(f)
        test.update({ds: nc_ds})
        stamp_fields.update({ds: ['lon', 'lat', 'plant_water_availability_east']})
        nc_ds.close()

    display = act.plotting.ContourDisplay(test, figsize=(8, 8))
    display.create_contour(fields=stamp_fields, time=time, levels=50, alpha=0.5, twod_dim_value=5)

    try:
        return display.fig
    finally:
        matplotlib.pyplot.close(display.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_contour2():
    files = sample_files.EXAMPLE_MET_CONTOUR
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        ds = act.io.arm.read_arm_netcdf(f)
        data.update({f: ds})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update({f: ['lon', 'lat', 'atmos_pressure']})

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50, contour='contour', cmap='viridis')
    display.plot_vectors_from_spd_dir(
        fields=wind_fields, time=time, mesh=False, grid_delta=(0.1, 0.1)
    )
    display.plot_station(fields=station_fields, time=time, markersize=7, color='pink')

    try:
        return display.fig
    finally:
        matplotlib.pyplot.close(display.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_contourf():
    files = sample_files.EXAMPLE_MET_CONTOUR
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        ds = act.io.arm.read_arm_netcdf(f)
        data.update({f: ds})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update(
            {
                f: [
                    'lon',
                    'lat',
                    'atmos_pressure',
                    'temp_mean',
                    'rh_mean',
                    'vapor_pressure_mean',
                    'temp_std',
                ]
            }
        )

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50, contour='contourf', cmap='viridis')
    display.plot_vectors_from_spd_dir(
        fields=wind_fields, time=time, mesh=True, grid_delta=(0.1, 0.1)
    )
    display.plot_station(fields=station_fields, time=time, markersize=7, color='red')

    try:
        return display.fig
    finally:
        matplotlib.pyplot.close(display.fig)


@pytest.mark.mpl_image_compare(tolerance=10)
def test_contourf2():
    files = sample_files.EXAMPLE_MET_CONTOUR
    time = '2019-05-08T04:00:00.000000000'
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        ds = act.io.arm.read_arm_netcdf(f)
        data.update({f: ds})
        fields.update({f: ['lon', 'lat', 'temp_mean']})
        wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
        station_fields.update(
            {
                f: [
                    'lon',
                    'lat',
                    'atmos_pressure',
                    'temp_mean',
                    'rh_mean',
                    'vapor_pressure_mean',
                    'temp_std',
                ]
            }
        )

    display = ContourDisplay(data, figsize=(8, 8))
    display.create_contour(fields=fields, time=time, levels=50, contour='contourf', cmap='viridis')
    display.plot_vectors_from_spd_dir(
        fields=wind_fields, time=time, mesh=False, grid_delta=(0.1, 0.1)
    )
    display.plot_station(fields=station_fields, time=time, markersize=7, color='pink')

    try:
        return display.fig
    finally:
        matplotlib.pyplot.close(display.fig)



"""
Cloud Base Height Retrievals
----------------------------

This example shows how to calculate the cloud base heights
using the sobel edge detection method.  This can be used
for vertical radar and lidar data.

Author: Adam Theisen

"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt
import act
import numpy as np

# Read Ceilometer data for an example
filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_ceil)

ds = act.retrievals.cbh.generic_sobel_cbh(
    ds, variable='backscatter', height_dim='range', var_thresh=1000.0, fill_na=0.0
)

# Plot the cloud base height data
display = act.plotting.TimeSeriesDisplay(ds, subplot_shape=(1, 2), figsize=(16, 6))
display.plot('backscatter', subplot_index=(0, 0))
title = 'SGP Ceilometer with Lidar-Calculated CBH Overplotted'
display.plot('first_cbh', subplot_index=(0, 0), color='k', set_title=title)

display.plot('backscatter', subplot_index=(0, 1))
title = 'SGP Ceilometer with CBH Overplotted'
display.plot('cbh_sobel_backscatter', color='k', subplot_index=(0, 1), set_title=title)

diff = ds['first_cbh'].values - ds['cbh_sobel_backscatter'].values

print("Average difference between ceilomter and sobel heights ", np.nanmean(diff))

ds.close()
plt.show()
